package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.client.Minecraft;
import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.StringTag;
import net.minecraft.nbt.Tag;
import net.minecraft.network.Connection;
import net.minecraft.network.protocol.PacketFlow;
import net.minecraft.network.protocol.game.ClientboundBlockEntityDataPacket;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.item.ItemEntity;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.level.BlockGetter;
import net.minecraft.world.level.Explosion;
import net.minecraft.world.level.ExplosionDamageCalculator;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.Blocks;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.world.level.levelgen.structure.BoundingBox;
import net.minecraft.world.level.material.FluidState;
import net.minecraft.world.level.saveddata.SavedData;
import net.minecraft.world.phys.AABB;
import net.minecraft.world.phys.Vec3;
import net.minecraftforge.event.TickEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Optional;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityBlackHole extends BlockEntity {
	public BlockEntityBlackHole(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.BLACK_HOLE, p_155229_, p_155230_);
		setMass(mass);
	}

	public static final Fraction NORMAL_MASS = new Fraction(1_000);
	private Fraction mass = NORMAL_MASS;
	private double gravitation_strength = 1;
	private boolean suppress_updates = true;
	public boolean spin_direction = false;

	public double getGravitationStrength() {
		return gravitation_strength;
	}

	private static class BlackHoleExplosionDamageCalculator extends ExplosionDamageCalculator {
		@Override
		public Optional<Float> getBlockExplosionResistance(Explosion explosion, BlockGetter getter, BlockPos pos, BlockState state, FluidState fstate) {
			return state.isAir() && fstate.isEmpty() ?
				Optional.empty() :
				Optional.of(
					(float) Math.sqrt(
						Math.max(0,
							Math.max(
								state.getExplosionResistance(getter, pos, explosion),
								fstate.getExplosionResistance(getter, pos, explosion)
							)
						)
					)
				);
		}

		private static final BlackHoleExplosionDamageCalculator INSTANCE = new BlackHoleExplosionDamageCalculator();
	}

	public static class BlackHoleExplosionQueue extends SavedData {
		private record QueuedExplosion(double x, double y, double z, float radius) {
			public CompoundTag serializeNBT() {
				var tag = new CompoundTag();

				tag.putDouble("x", x);
				tag.putDouble("y", y);
				tag.putDouble("z", z);
				tag.putFloat("radius", radius);

				return tag;
			}

			public static QueuedExplosion deserializeNBT(CompoundTag tag) {
				return new QueuedExplosion(tag.getDouble("x"), tag.getDouble("y"), tag.getDouble("z"), tag.getFloat("radius"));
			}

			public void explode(Level level) {
				level.explode(
					null,
					Registry.DAMAGE_HAWKING_RADIATION,
					BlackHoleExplosionDamageCalculator.INSTANCE,
					x,
					y,
					z,
					radius,
					false,
					Explosion.BlockInteraction.DESTROY);
			}
		}

		private record RingExplosion(double x, double y, double z, double radius, float strength) {
			public CompoundTag serializeNBT() {
				var tag = new CompoundTag();

				tag.putDouble("x", x);
				tag.putDouble("y", y);
				tag.putDouble("z", z);
				tag.putDouble("radius", radius);

				return tag;
			}

			public static RingExplosion deserializeNBT(CompoundTag tag) {
				return new RingExplosion(tag.getDouble("x"), tag.getDouble("y"), tag.getDouble("z"), tag.getDouble("radius"), tag.getFloat("strength"));
			}

			public void explode(BlackHoleExplosionQueue queue) {
				final int fragments = (int) radius * 8;

				final double stack_step = Math.PI / (double) fragments;
				final double sector_step = Math.PI / (double) fragments * 2;

				for (int stack = 0; stack < fragments; stack++) {
					final double stack_angle = Math.PI / 2 - stack * stack_step;
					final double xy = radius * 15 * Math.cos(stack_angle);
					final double z = radius * 15 * Math.sin(stack_angle);

					for (int sector = 0; sector < fragments; sector++) {
						final double sector_angle = sector * sector_step;

						queue.explode(
							this.x + xy * Math.cos(sector_angle),
							this.y + xy * Math.sin(sector_angle),
							this.z + z,
							strength
						);
					}
				}
			}
		}

		private int index = 0;
		private int index_ring = 0;
		private final ArrayList<QueuedExplosion> explosions = new ArrayList<>();
		private final ArrayList<RingExplosion> rings = new ArrayList<>();
		private final ServerLevel level;

		public BlackHoleExplosionQueue(ServerLevel level) {
			this.level = level;
		}

		@Override
		public CompoundTag save(CompoundTag tag) {
			var list = new ListTag();
			var list_rings = new ListTag();

			for (int i = index; i < explosions.size(); i++)
				list.add(explosions.get(i).serializeNBT());

			for (int i = index_ring; i < rings.size(); i++)
				list_rings.add(rings.get(i).serializeNBT());

			tag.put("explosions", list);
			tag.put("rings", list_rings);

			return tag;
		}

		public void load(CompoundTag tag) {
			explosions.clear();
			rings.clear();
			index = 0;
			index_ring = 0;

			for (var explosion : tag.getList("explosions", Tag.TAG_COMPOUND))
				explosions.add(QueuedExplosion.deserializeNBT((CompoundTag) explosion));

			for (var ring : tag.getList("rings", Tag.TAG_COMPOUND))
				rings.add(RingExplosion.deserializeNBT((CompoundTag) ring));
		}

		public static BlackHoleExplosionQueue factory(ServerLevel level) {
			return level.getDataStorage().computeIfAbsent(
				(tag) -> {
					var factory = new BlackHoleExplosionQueue(level);
					factory.load(tag);
					return factory;
				},

				() -> new BlackHoleExplosionQueue(level),

				"otm_blackhole_explosion_queue"
			);
		}

		public void explode(double x, double y, double z, float radius) {
			explosions.add(new QueuedExplosion(x, y, z, radius));
			setDirty();
		}

		public void explodeRing(double x, double y, double z, double radius, float strength) {
			rings.add(new RingExplosion(x, y, z, radius, strength));
			setDirty();
		}

		public void tick() {
			if (explosions.size() != 0) {
				setDirty();

				int iterations = 0;

				for (int i = index; i < explosions.size(); i++) {
					explosions.get(i).explode(level);
					index++;

					if (iterations++ == 4) {
						break;
					}
				}

				if (index >= explosions.size()) {
					index = 0;
					explosions.clear();
				}
			} else if (rings.size() != 0) {
				if (index_ring >= rings.size()) {
					index_ring = 0;
					rings.clear();
				} else {
					rings.get(index_ring++).explode(this);
				}
			}
		}

		@SubscribeEvent
		public static void onWorldTick(TickEvent.WorldTickEvent event) {
			if (event.phase == TickEvent.Phase.START && event.world instanceof ServerLevel level) {
				factory(level).tick();
			}
		}
	}

	public void collapse() {
		level.setBlock(getBlockPos(), Blocks.AIR.defaultBlockState(), Block.UPDATE_ALL);

		if (gravitation_strength > 0.25) {
			final double x0 = getBlockPos().getX() + 0.5;
			final double y0 = getBlockPos().getY() + 0.5;
			final double z0 = getBlockPos().getZ() + 0.5;

			var queue = BlackHoleExplosionQueue.factory((ServerLevel) level);

			for (int radius = 0; radius < Math.ceil(gravitation_strength * 4); radius++) {
				queue.explodeRing(
					x0,
					y0,
					z0,
					radius,
					(float) Math.min(20, Math.max(1, (gravitation_strength * 4 - radius) * 20)));
			}
		} else {
			level.explode(
				null,
				Registry.DAMAGE_HAWKING_RADIATION,
				null,
				getBlockPos().getX() + 0.5D,
				getBlockPos().getY() + 0.5D,
				getBlockPos().getZ() + 0.5D,
				(float) gravitation_strength * 60,
				false,
				Explosion.BlockInteraction.DESTROY);
		}
	}

	public void addMass(Fraction mass) {
		setMass(this.mass.plus(mass));
	}

	public void setMass(Fraction mass) {
		if (mass.compareTo(Fraction.ZERO) <= 0) {
			collapse();
			return;
		}

		this.mass = mass;
		setChanged();
		gravitation_strength = mass.div(NORMAL_MASS).toDouble();

		if (gravitation_strength > 1) {
			gravitation_strength = Math.min(20, Math.sqrt(gravitation_strength));
		} else {
			gravitation_strength = Math.max(0.08, Math.pow(gravitation_strength, 2));
		}

		if (level != null && !level.isClientSide && !suppress_updates)
			level.sendBlockUpdated(getBlockPos(), getBlockState(), getBlockState(), Block.UPDATE_CLIENTS);

		affected_bounds = new BoundingBox(
			(int) (-30 * gravitation_strength),
			(int) (-30 * gravitation_strength),
			(int) (-30 * gravitation_strength),
			(int) (30 * gravitation_strength),
			(int) (30 * gravitation_strength),
			(int) (30 * gravitation_strength)
		).move(getBlockPos());

		affected_bounds_aabb = AABB.of(affected_bounds);
	}

	// shared functions
	public CompoundTag writeBlackHoleData(CompoundTag tag) {
		tag.putString("mass", mass.toString());
		tag.putBoolean("spin_direction", spin_direction);
		return tag;
	}

	public void readBlackHoleData(CompoundTag tag) {
		if (tag.contains("mass"))
			setMass(Fraction.deserializeNBT(tag.get("mass")));

		spin_direction = tag.getBoolean("spin_direction");
	}

	// disk io
	@Override
	public void saveAdditional(CompoundTag nbt) {
		super.saveAdditional(nbt);
		writeBlackHoleData(nbt);
	}

	@Override
	public void load(CompoundTag tag) {
		super.load(tag);
		readBlackHoleData(tag);
	}

	// received either by game engine (from getUpdateTag on ClientLevelPacket)
	// or by onDataPacket
	@Override
	public void handleUpdateTag(CompoundTag tag) {
		super.handleUpdateTag(tag);
		readBlackHoleData(tag);
	}

	// called by game engine for ClientLevelPacket
	@Override
	public CompoundTag getUpdateTag() {
		return writeBlackHoleData(super.getUpdateTag());
	}

	// called by game engine on block updates
	@Nullable
	@Override
	public ClientboundBlockEntityDataPacket getUpdatePacket() {
		return ClientboundBlockEntityDataPacket.create(this);
	}

	// called by game engine by forge patches when ClientboundBlockEntityDataPacket is received
	@Override
	public void onDataPacket(Connection net, ClientboundBlockEntityDataPacket pkt) {
		super.onDataPacket(net, pkt);

		if (net.getReceiving() == PacketFlow.CLIENTBOUND) {
			handleUpdateTag(pkt.getTag());
		}
	}

	private BoundingBox affected_bounds;
	private AABB affected_bounds_aabb;

	public BoundingBox affectedBounds() {
		return affected_bounds;
	}

	public AABB affectedBoundsAABB() {
		return affected_bounds_aabb;
	}

	private void setDeltaMovement(Entity living, Vec3 center, double distance, boolean weaker) {
		//final double mult = Math.min(2, (30 * this.gravitation_strength) / Math.max(1, Math.pow(distance, 2)));
		// Сила притяжения
		final double mult = Math.pow(1 - distance / (30 * this.gravitation_strength), 2) * this.gravitation_strength / 8;

		// Притяжение к ядру
		final var delta = living.position().vectorTo(center).normalize();

		if (distance < this.gravitation_strength) {
			living.setDeltaMovement(living.getDeltaMovement().add(delta.multiply(mult, mult, mult)));
		} else {
			// Закручивание
			final var rotate = spin_direction ? delta.yRot((float) (Math.PI / 2)) : delta.yRot((float) (-Math.PI / 2));

			if (weaker)
				living.setDeltaMovement(living.getDeltaMovement().add(rotate.multiply(mult * 0.4f, mult * 0.4f, mult * 0.4f)).add(delta.multiply(mult * 0.33f, mult * 0.33f, mult * 0.33f)));
			else
				living.setDeltaMovement(living.getDeltaMovement().add(rotate.multiply(mult * 0.4f, mult * 0.4f, mult * 0.4f)).add(delta.multiply(mult, mult, mult)));
		}
	}

	public static <T extends BlockEntity> void clientTicker(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityBlackHole tile) {
			var ply = Minecraft.getInstance().player;
			final var center = Vec3.atCenterOf(tile.getBlockPos());

			if (!ply.getAbilities().mayfly) {
				final double distance = ply.position().distanceTo(center);
				tile.setDeltaMovement(ply, center, distance, true);
			}

			for (var item : tile.level.getEntitiesOfClass(ItemEntity.class, tile.affected_bounds_aabb)) {
				final double distance = item.position().distanceTo(center);
				tile.setDeltaMovement(item, center, distance, false);
			}
		}
	}

	public static final Fraction HAWKING_MASS_LOSE_STEP = new Fraction("-0.1");

	public static <T extends BlockEntity> void ticker(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityBlackHole tile) {
			tile.suppress_updates = false;

			final var center = Vec3.atCenterOf(tile.getBlockPos());

			for (var living : level.getEntitiesOfClass(LivingEntity.class, tile.affected_bounds_aabb)) {
				final double distance = living.position().distanceTo(center);

				if (!(living instanceof Player ply) || !ply.getAbilities().mayfly) {
					tile.setDeltaMovement(living, center, distance, false);
				}

				if (distance < tile.gravitation_strength + 1) {
					living.hurt(Registry.DAMAGE_EVENT_HORIZON, (float) (tile.gravitation_strength / distance));
				}
			}

			for (var item : level.getEntitiesOfClass(ItemEntity.class, tile.affected_bounds_aabb)) {
				final double distance = item.position().distanceTo(center);
				tile.setDeltaMovement(item, center, distance, false);

				if (distance < tile.gravitation_strength + 1) {
					if (item.hurt(Registry.DAMAGE_EVENT_HORIZON, (float) (tile.gravitation_strength / distance)) && item.isRemoved()) {
						if (item.getItem().getItem() == Registry.Items.GRAVITATIONAL_DISRUPTOR) {
							tile.collapse();
						} else {
							var mass = MatterRegistry.getMatterValue(item.getItem());

							if (mass.compareTo(Fraction.ZERO) > 0)
								tile.addMass(mass);
						}
					}
				}
			}

			// шанс 1% что черная дыра потеряет 0.1 MtU каждую секунду * силу гравитации дыры ^ -1
			if (level.random.nextDouble() < 0.01 * 0.05 * (1 / tile.gravitation_strength)) {
				tile.addMass(HAWKING_MASS_LOSE_STEP);
			}
		}
	}
}
