package ru.dbotthepony.mc.otm.storage;

import net.minecraft.MethodsReturnNonnullByDefault;
import ru.dbotthepony.mc.otm.core.Fraction;

import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;
import java.util.function.Supplier;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class VirtualComponent<T extends IStorageStack> implements IStorageComponent<T>, IStorageListener<T> {
	public record LocalTuple<T extends IStorageStack>(T stack, UUID id, List<RemoteTuple<T>> tuples) implements IStorageTuple<T> {
		@Override
		public boolean equals(Object obj) {
			return obj instanceof LocalTuple tuple && tuple.id.equals(id) || obj instanceof UUID id && this.id.equals(id);
		}

		@Override
		public int hashCode() {
			return id.hashCode();
		}
	}

	public record RemoteTuple<T extends IStorageStack>(T object, UUID remote_id, IStorageView<T> provider, LocalTuple<T> local) {
		public T extract(Fraction amount, boolean simulate) {
			return provider.extractStack(remote_id, amount, simulate);
		}

		public Fraction extractCount(Fraction amount, boolean simulate) {
			return provider.extractStackCount(remote_id, amount, simulate);
		}

		@Override
		public boolean equals(Object obj) {
			return obj instanceof RemoteTuple tuple && tuple.remote_id.equals(remote_id) || obj instanceof UUID id && id.equals(remote_id);
		}

		@Override
		public int hashCode() {
			return remote_id.hashCode();
		}
	}

	protected final StorageObjectTuple<T> identity;

	@Override
	public Class<T> storageIdentity() {
		return identity.identity();
	}

	public VirtualComponent(Class<T> identity) {
		this.identity = StorageObjectRegistry.getOrError(identity);
	}

	public static <T extends IStorageStack> Supplier<VirtualComponent<T>> factory(Class<T> identity) {
		return () -> new VirtualComponent<>(identity);
	}

	// удаленный UUID -> Кортеж
	protected final HashMap<UUID, RemoteTuple<T>> indexed_by_remote_uuid = new HashMap<>();

	// локальный UUID -> Локальный кортеж
	protected final HashMap<UUID, LocalTuple<T>> indexed_by_local_uuid = new HashMap<>();

	// Хеш ключ -> Список Локальных кортежей
	protected final HashMap<Object, ArrayList<LocalTuple<T>>> partitions = new HashMap<>();

	// ArrayList для скорости работы
	protected final ArrayList<IStorageListener<T>> listeners = new ArrayList<>();
	protected final ArrayList<IStorageConsumer<T>> consumers = new ArrayList<>();

	public void add(IStorageIdentity<T> identity) {
		if (identity instanceof IStorageView<T> provider) {
			provider.addListenerAuto(this);
		} else if (identity instanceof IStorageTrigger<T> provider) {
			provider.addListener(this);
		}

		if (identity instanceof IStorageConsumer<T> provider && !consumers.contains(provider)) {
			consumers.add(provider);
		}
	}

	public void remove(IStorageIdentity<T> identity) {
		if (identity instanceof IStorageView<T> provider) {
			provider.removeListenerAuto(this);
		} else if (identity instanceof IStorageTrigger<T> provider) {
			provider.removeListener(this);
		}

		if (identity instanceof IStorageConsumer<T> provider) {
			consumers.remove(provider);
		}
	}

	@Override
	public T insertStack(T stack, boolean simulate) {
		var leftover = stack;

		for (var consumer : consumers) {
			leftover = consumer.insertStack(leftover, simulate);

			if (leftover.isEmpty()) {
				return leftover;
			}
		}

		return leftover;
	}

	@Override
	public boolean addListener(IStorageListener<T> listener) {
		if (!listeners.contains(listener)) {
			listeners.add(listener);
			return true;
		}

		return false;
	}

	@Override
	public boolean removeListener(IStorageListener<T> listener) {
		return listeners.remove(listener);
	}

	@Override
	public T getStack(UUID id) {
		final var tuple = indexed_by_local_uuid.get(id);
		return tuple != null ? tuple.stack : identity.empty();
	}

	@Override
	@SuppressWarnings("unchecked")
	public T extractStack(UUID id, Fraction amount, boolean simulate) {
		var tuple = indexed_by_local_uuid.get(id);

		if (tuple == null || amount.compareTo(Fraction.ZERO) == 0)
			return identity.empty();

		if (amount.compareTo(Fraction.MINUS_ONE) <= 0)
			amount = tuple.stack.getMaxStackSize().orElse(tuple.stack.getCount());

		var extract = tuple.stack.getCount().min(amount);
		var extracted = Fraction.ZERO;

		final var copy = (T) tuple.stack.copy();

		for (var remote_tuple : tuple.tuples) {
			extracted = extracted.plus(remote_tuple.extractCount(extract.minus(extracted), simulate));

			if (extracted.compareTo(extract) >= 0)
				break;
		}

		if (extracted.compareTo(Fraction.ZERO) > 0) {
			copy.setCount(extracted);
			return copy;
		}

		return identity.empty();
	}

	@Override
	public List<IStorageTuple<T>> getStacks() {
		int capacity = 0;

		for (var list : partitions.values())
			capacity += list.size();

		final var output = new ArrayList<IStorageTuple<T>>(capacity);

		for (var listing : partitions.values())
			output.addAll(listing);

		return output;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void addObject(T stack, UUID id, IStorageView<T> provider) {
		if (indexed_by_remote_uuid.containsKey(id)) {
			throw new IllegalStateException("Already tracking tuple with id " + id);
		}

		var items = partitions.computeIfAbsent(stack.partitionKey(), (k) -> new ArrayList<>());
		LocalTuple<T> local_tuple = null;

		for (var item : items) {
			if (item.stack.sameItem(stack)) {
				item.stack.grow(stack.getCount());
				local_tuple = item;
				break;
			}
		}

		boolean added = local_tuple == null;

		if (added) {
			local_tuple = new LocalTuple<>((T) stack.copy(), UUID.randomUUID(), new ArrayList<>(1));
			items.add(local_tuple);
			indexed_by_local_uuid.put(local_tuple.id, local_tuple);
		}

		var tuple = new RemoteTuple<>((T) stack.copy(), id, provider, local_tuple);
		local_tuple.tuples.add(tuple);
		indexed_by_remote_uuid.put(id, tuple);

		if (added) {
			for (var listener : listeners) {
				listener.addObject(local_tuple.stack, local_tuple.id, this);
			}
		} else {
			for (var listener : listeners) {
				listener.changeObject(local_tuple.id, local_tuple.stack.getCount());
			}
		}
	}


	@Override
	public void changeObject(UUID id, Fraction new_count) {
		assert new_count.compareTo(Fraction.ZERO) > 0;
		var tuple = indexed_by_remote_uuid.get(id);

		if (tuple == null)
			throw new IllegalStateException("No such tuple with id " + id);

		final var diff = new_count.minus(tuple.object.getCount());
		tuple.object.setCount(new_count);
		tuple.local.stack.grow(diff);

		for (var listener : listeners) {
			listener.changeObject(tuple.local.id, tuple.local.stack.getCount());
		}
	}

	@Override
	public void removeObject(UUID id) {
		var tuple = indexed_by_remote_uuid.get(id);

		if (tuple == null)
			throw new IllegalStateException("No such tuple with id " + id);

		final var item = tuple.local.stack.partitionKey();
		tuple.local.stack.shrink(tuple.object.getCount());
		tuple.local.tuples.remove(tuple);

		indexed_by_remote_uuid.remove(id);

		final boolean a = tuple.local.stack.getCount().compareTo(Fraction.ZERO) <= 0;
		final boolean b = tuple.local.tuples.size() == 0;

		if (a || b) {
			if (!(a && b))
				throw new IllegalStateException("View object is empty, but tuple list is not!");

			indexed_by_local_uuid.remove(tuple.local.id);
			partitions.get(item).remove(tuple.local);

			for (var listener : listeners) {
				listener.removeObject(tuple.local.id);
			}
		}
	}
}
