package ru.dbotthepony.mc.otm.capability.drive

import java.util.HashMap
import java.util.UUID
import net.minecraft.ReportedException
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtIo
import net.minecraft.CrashReport
import net.minecraft.server.MinecraftServer
import java.util.concurrent.locks.LockSupport
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.eventbus.api.EventPriority
import net.minecraftforge.event.TickEvent.ServerTickEvent
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.server.ServerAboutToStartEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import net.minecraftforge.event.world.WorldEvent
import org.apache.logging.log4j.LogManager
import java.io.File
import java.lang.ref.WeakReference
import java.util.ArrayList

/**
 * Let me answer everyone's questions about:
 *
 * Why?
 *
 * There are several reasons:
 * 0. This data can get very large very quickly, even when playing singleplayer (and much quicker and bigger when hosting a dedicated server)
 * 1. This data can not be stored inside ItemStack.ForgeCaps due to it's size
 * 2. This data can not be stored inside unshared (server only) nbt tag because, again, mods prone to use and interact with
 * it wrong, causing loss of stored data or mods exposing full content of a drive inside their own tag (which cause very real "NBT size too large"
 * network kicks, often locking players out of server/singleplayer worlds
 * 3. net.minecraft.world.level.saveddata.SaveData is for storing everything inside one dat file, which
 * is performance tanking, because we have to write *entire* NBT on each save, not the data of Drives that are dirty
 */
@Suppress("unused")
object DrivePool {
	private val LOGGER = LogManager.getLogger()
	private val pool = HashMap<UUID, WeakDriveReference>()
	private var thread: Thread? = null
	private var stopping = false
	private var exception: ReportedException? = null
	private const val DATA_PATH = "data/otm_drives"

	@JvmStatic
	operator fun <T : IMatteryDrive<*>> get(
		id: UUID,
		factory_stored: (CompoundTag) -> T,
		factory_empty: () -> T
	): T {
		if (!isLegalAccess())
			throw IllegalAccessException("Can not access drive pool from outside of server thread.")

		val get = pool[id]

		if (get != null) {
			val accessed = get.access()

			if (accessed != null) {
				return accessed as T
			}
		}

		val file = File(baseDirectory, "$id.dat")

		if (file.exists()) {
			try {
				val factorize = factory_stored(NbtIo.readCompressed(file))
				pool[id] = WeakDriveReference(factorize)
				return factorize
			} catch (error: Throwable) {
				val report = CrashReport("Reading OTM Drive from disk", error)
				val category = report.addCategory("Corrupt drive")
				category.setDetail("UUID", id)
				throw ReportedException(report)
			}
		}

		val factorize = factory_empty()
		pool[id] = WeakDriveReference(factorize)
		return factorize
	}

	@JvmStatic
	fun put(id: UUID, drive: IMatteryDrive<*>) {
		if (!isLegalAccess())
			throw IllegalAccessException("Can not access drive pool from outside of server thread.")

		pool[id] = WeakDriveReference(drive)
	}

	@JvmStatic
	fun markDirty(id: UUID) {
		if (!isLegalAccess())
			throw IllegalAccessException("Can not access drive pool from outside of server thread.")

		pool[id]?.access()
	}

	private var backlog = ArrayList<BacklogLine>()
	private var baseDirectory: File? = null
	private var serverThread: Thread? = null

	private fun thread(server: MinecraftServer) {
		while (true) {
			LockSupport.park()

			if (stopping) {
				return
			}

			try {
				// TODO: Syncing status with main thread, since server stop can occur during work.
				sync()
			} catch (error: ReportedException) {
				exception = error
				return
			}
		}
	}

	@SubscribeEvent(priority = EventPriority.LOWEST)
	fun onServerPostTick(event: ServerTickEvent) {
		if (event.phase == TickEvent.Phase.END) {
			if (exception != null) {
				throw exception!!
			}
		}
	}

	/**
	 * Returns whenever running on server thread. Calling [get], [put] or [markDirty] will throw an exception if this returns false.
	 *
	 * If you are making your own drive for your own storage stack, feed code outside of server thread (e.g. client code) dummy disk.
	 */
	@JvmStatic
	fun isLegalAccess(): Boolean {
		return serverThread == null || Thread.currentThread() === serverThread
	}

	@SubscribeEvent(priority = EventPriority.HIGHEST)
	fun serverStartEvent(event: ServerAboutToStartEvent) {
		if (thread != null && thread!!.isAlive) {
			LOGGER.error("FMLServerStartedEvent fired twice.")
			LOGGER.error("Attempting to start another DrivePool I/O thread when already running one!")
			return
		}

		pool.clear()

		val server = event.server
		baseDirectory = File(server.storageSource.worldDir.toFile(), DATA_PATH)
		baseDirectory!!.mkdirs()

		serverThread = Thread.currentThread()

		stopping = false
		thread = Thread(null, { thread(server) }, "Overdrive That Matters DrivePool IO")
		thread!!.start()
	}

	@SubscribeEvent(priority = EventPriority.HIGHEST)
	fun serverStopEvent(event: ServerStoppingEvent) {
		val thread = thread

		if (thread != null && thread.isAlive) {
			stopping = true
			LockSupport.unpark(thread)
		}

		if (exception == null) {
			writeBacklog()
			sync()
		}

		pool.clear()
	}

	@SubscribeEvent
	fun onWorldSave(event: WorldEvent.Save) {
		writeBacklog()
	}

	private fun writeBacklog() {
		var needsSync = false
		var removeKeys: ArrayList<UUID>? = null

		for ((key, value) in pool) {
			val tag = value.sync()

			if (tag != null) {
				LOGGER.info("Serializing OTM Drive {}", key)
				val index = backlog.indexOf<Any>(key)

				if (index != -1) {
					backlog[index] = BacklogLine(key, tag, value.drive()!!)
				} else {
					backlog.add(BacklogLine(key, tag, value.drive()!!))
				}

				needsSync = true
			} else if (!value.valid()) {
				if (removeKeys == null)
					removeKeys = ArrayList()

				removeKeys.add(key)
			}
		}

		if (removeKeys != null) {
			for (key in removeKeys) {
				pool.remove(key)
			}
		}

		if (needsSync) {
			LockSupport.unpark(thread)
		}
	}

	private fun sync() {
		val copy = backlog
		backlog = ArrayList()

		for (entry in copy) {
			try {
				LOGGER.info("Syncing OTM Drive {}", entry.file)

				val oldFile = File(baseDirectory, entry.file.toString() + ".dat_old")

				if (oldFile.exists()) {
					check(oldFile.delete()) { "Unable to delete old dat file" }
				}

				val newFile = File(baseDirectory, entry.file.toString() + ".dat")

				if (newFile.exists()) {
					check(newFile.renameTo(oldFile)) { "Unable to move old dat file" }
				}

				NbtIo.writeCompressed(entry.tag, newFile)
			} catch (error: Throwable) {
				val report = CrashReport("Syncing OTM Drives state to disk", error)
				val category = report.addCategory("Corrupt drive")
				category.setDetail("UUID", entry.file.toString())
				category.setDetail("Stored item count", entry.drive.storedCount)
				category.setDetail("Capacity", entry.drive.driveCapacity)
				throw ReportedException(report)
			}
		}
	}
}

private class WeakDriveReference(drive: IMatteryDrive<*>) {
	private var drive: IMatteryDrive<*>? = drive
	private val weak = WeakReference(drive)

	fun valid(): Boolean {
		return drive != null || weak.get() != null
	}

	fun drive(): IMatteryDrive<*>? {
		return drive ?: weak.get()
	}

	fun sync(): CompoundTag? {
		var drive = drive

		if (drive == null) {
			drive = weak.get()

			if (drive == null)
				return null
		}

		if (!drive.isDirty) {
			this.drive = null
			return null
		}

		val tag = drive.serializeNBT()
		drive.isDirty = false
		this.drive = null

		return tag
	}

	fun access(): IMatteryDrive<*>? {
		this.drive = weak.get()
		return drive
	}
}

private data class BacklogLine(val file: UUID, val tag: CompoundTag, val drive: IMatteryDrive<*>) {
	override fun equals(other: Any?): Boolean {
		if (other is BacklogLine) {
			return other.file == file
		}

		if (other is UUID) {
			return other == file
		}

		return false
	}
}