package ru.dbotthepony.mc.otm.storage

import ru.dbotthepony.mc.otm.core.Fraction
import java.util.*

interface IStorageStack {
	fun copy(): IStorageStack

	var count: Fraction
	fun isEmpty(): Boolean

	/**
	 * @return max stack size for this stack object
	 * [Optional.empty()] if unlimited (default)
	 */
	fun getMaxStackSize(): Optional<Fraction> = Optional.empty()

	/**
	 * Returns Identity utilized to partition view table (if it has any).
	 * Defaults to no partitioning; meaning performance will degrade much quicker than it should.
	 * Good object is an object that will influence on [sameItem] return result, making it return true.
	 * It is strictly considered that if !this.itemIdentity().equals(other.itemIdentity);
	 * then this.sameItem(other) will never return true.
	 *
	 * If implemented, and storage is also partitioned properly, then performance will be flat equally to view table's performance.
	 *
	 * Example: ItemStack#item
	 */
	fun partitionKey(): Any {
		return Any::class.java
	}

	/**
	 * [other] other object to compare; if not the same of this object type just return false
	 * Returns boolean representing whenever internal state of [this] equals to [other];
	 * this include tags, mapping IDs, capabilities, **excluding** amount;
	 * behavior is pretty much the same as ItemStack.isSameItemSameTags
	 */
	fun sameItem(other: IStorageStack): Boolean

	fun grow(amount: Fraction) {
		count += amount
	}

	fun shrink(amount: Fraction) {
		count -= amount
	}
}

/**
 * Storage system root, along IStorageStack interface
 */
interface IStorageIdentity<T : IStorageStack> {
	fun storageIdentity(): Class<T>
}

interface IStorageTrigger<T : IStorageStack> : IStorageIdentity<T> {
	/**
	 * [listener] is [IStorageListener] which want to subscribe to our events
	 */
	fun addListener(listener: IStorageListener<T>): Boolean

	/**
	 * [listener] is [IStorageListener] which want to unsubscribe from our events
	 */
	fun removeListener(listener: IStorageListener<T>): Boolean
}

interface IStorageConsumer<T : IStorageStack> : IStorageIdentity<T> {
	fun insertStack(stack: T, simulate: Boolean): T
}

interface IStorageView<T : IStorageStack> : IStorageTrigger<T> {
	/**
	 * @param id identifier of object
	 * @return stored object (not a copy). Do not edit it.
	 */
	fun getStack(id: UUID): T

	/**
	 * @param id identifier of object to extract
	 * @param amount amount of units to extract
	 * @param simulate whenever to simulate the action or not
	 * @return copy of object, with amount of units actually extracted
	 */
	fun extractStack(id: UUID, amount: Fraction, simulate: Boolean): T

	/**
	 * Designed for views, for extraction with less computation overhead caused by
	 * copying object extracted
	 *
	 * @param id identifier of object to extract
	 * @param amount desired amount to extract
	 * @param simulate whenever to simulate the action or not
	 * @return amount extracted
	 */
	fun extractStackCount(id: UUID, amount: Fraction, simulate: Boolean): Fraction? {
		return extractStack(id, amount, simulate).count
	}

	fun getStacks(): Collection<IStorageTuple<T>>

	fun addListenerAuto(listener: IStorageListener<T>): Boolean {
		if (addListener(listener)) {
			for (stack in getStacks()) {
				listener.addObject(stack.stack(), stack.id(), this)
			}

			return true
		}

		return false
	}

	fun removeListenerAuto(listener: IStorageListener<T>): Boolean {
		if (removeListener(listener)) {
			for (stack in getStacks()) {
				listener.removeObject(stack.id())
			}

			return true
		}

		return false
	}
}

interface IStorageListener<T : IStorageStack> {
	/**
	 * Fired on whenever an object is added (to listener) we subscribed to
	 */
	fun addObject(stack: T, id: UUID, provider: IStorageView<T>)

	/**
	 * Fired on whenever an object is changes on listener we subscribed to
	 */
	fun changeObject(id: UUID, new_count: Fraction)

	/**
	 * Fired on whenever an object is removed from listener we subscribed to
	 */
	fun removeObject(id: UUID)
}

interface IStorageTuple<T : IStorageStack> {
	val id: UUID
		get() = id()

	val stack: T
		get() = stack()

	fun id(): UUID
	fun stack(): T
}

data class StorageTuple<T : IStorageStack>(override val id: UUID, override val stack: T) : IStorageTuple<T> {
	override fun id(): UUID {
		return id
	}

	override fun stack(): T {
		return stack
	}
}

interface IStorageComponent<T : IStorageStack> : IStorageView<T>, IStorageConsumer<T>
